
/* drivers/video/ssd1906fb.c
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file COPYING in the main directory of this archive for
 * more details.
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/delay.h>

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/fb.h>
#include <linux/vmalloc.h>

#include <linux/jiffies.h>
#include <linux/timer.h>
#include <asm/atomic.h>

#include <asm/types.h>
#include <asm/io.h>

#include <mach/hardware.h>
#include <mach/at91_pio.h>
#include <mach/gpio.h>

#include "int070atft-fb.h"
#include "ssd1963_cmd.h"

// -----------------------------------------------------------------------------
// aktivujeme devel hlasky
#define DEBUG

// -----------------------------------------------------------------------------
enum
{
	DrvRemove = 0,

	Failed_FbRegister,
	Failed_CmapAlloc,
	Failed_VmemrefAlloc,
	Failed_VmemAlloc,
	Failed_IoMap,
	Failed_ReqRegion,
	Failed_FbAlloc,
};

// -----------------------------------------------------------------------------
// privatni parametry ovladace
struct int070atft_par_t
{
	// video pamet, ktera nahrazuje framebuffer v radici SSd1963, ktery neni
	// primo pristupny v adresovem prostoru CPU
	void *				vram_ptr;
	ulong		 		vram_size;
	// referencni video pamet
	void *				vram_ref_ptr;
	ulong				vram_ref_size;
	// casovac pro refresh lcd
	struct timer_list	refresh_timer;
	// (1 = deaktivuj casovac, 0 = casovac byl aktivovan)
	atomic_t 			stop_refresh_timer;
	// zacatek registru PIOC brany jako I/O pamet
	void __iomem *		pio_base;
	// povinna fiktivni paleta pro 16 barev a barvy kurzoru pro mod non-pallete
	u32					pseudo_palette[16];
};

// -----------------------------------------------------------------------------
// vychozi struktura fb_fix_screeninfo. V initu ovladace se zkopiruje do
// clenu fix struktury fb_info a pak uz se nepouziva.
static const struct fb_fix_screeninfo int070atftfb_fix __devinitdata =
{
	.id			= INT070ATFT_FBID,
	// packed pixel znamena, ze je pro dany pixel pouzito vice bitu, napr 16,
	// kde je zakodovana barva, napr. schema RGB565
	.type		= FB_TYPE_PACKED_PIXELS,
	// pouzivame truecolor
	.visual		= FB_VISUAL_TRUECOLOR,
	// panning je prohlizeni vetsi fotky na mensim displej - vypinam
	.xpanstep	= 0,
	.ypanstep	= 0,
	.ywrapstep	= 0,
	// HW akceleraci nemam
	.accel		= FB_ACCEL_NONE,

	/* vyplnuji se pozdeji ve funkci int070atftfb_probe:
	 * smem_start;			// Start of frame buffer mem
	 * smem_len;			// Length of frame buffer mem
	 * mmio_start;			// Start of Memory Mapped I/O
	 * mmio_len;			// Length of Memory Mapped I/O
	 * line_length;			// length of a line in bytes
	 */
};

// -----------------------------------------------------------------------------
// Pozn.: vychazim z toho, ze jedna instrukce pri 200MHz taktu trva 5ns
static void write_delay(void)
{
	// cca 15ns
	asm volatile
	(
		"NOP\n\t"
		"NOP\n\t"
		"NOP\n\t"
	);
}

static inline void read_delay(void)
{
	// cca 35ns
	asm volatile
	(
		"NOP\n\t"
		"NOP\n\t"
		"NOP\n\t"
		"NOP\n\t"
		"NOP\n\t"
		"NOP\n\t"
		"NOP\n\t"
	);
}

// -----------------------------------------------------------------------------
/*
 * Pozn.: Pozor neobsahuji pametove bariery (rmb, wmb, mb())!
 */
static inline unsigned int pio_readl(void __iomem * reg_addr)
{
	return __raw_readl(reg_addr);
}

static inline void pio_writel(void __iomem * reg_addr, unsigned int value)
{
	__raw_writel(value, reg_addr);
}

// -----------------------------------------------------------------------------
/*
 * prepina vodice 16bit sbernice pro TFT na cteni (PIO na input) nebo zapis
 * (PIO na vystup)
 */
static inline void set_bus_read(void __iomem * pio_base)
{
	pio_writel((pio_base + PIO_ODR), TFT_DB0_DB15_PIN_MASK);
}

static inline void set_bus_write(void __iomem * pio_base)
{
	pio_writel((pio_base + PIO_OER), TFT_DB0_DB15_PIN_MASK);
}

// -----------------------------------------------------------------------------
static inline unsigned int read_raw(void __iomem * pio_base)
{
	// orizneme na 16bit a posuneme zapisovanou hodnotu na PC16 - PC31
	unsigned int reg_val;

	// shodime RD
	at91_set_gpio_value(TFT_nRD, 0);

	// pockame min 36ns
	read_delay();

	// SSD1963 vystavi na D0-D15 data, vycteme si je
	reg_val = pio_readl(pio_base + PIO_PDSR);

	// ukoncime cteni dat vytazenim RD na log. 1
	at91_set_gpio_value(TFT_nRD, 1);

	return (reg_val >> 16);
}

// -----------------------------------------------------------------------------
static inline void write_raw(void __iomem * pio_base, unsigned int value)
{
	// orizneme na 16bit a posuneme zapisovanou hodnotu na PC16 - PC31
	value = (value & 0xFFFF) << 16;

	// shodime WR
	at91_set_gpio_value(TFT_nWR, 0);

	// na D0-D15 vystavime data
	pio_writel((pio_base + PIO_ODSR), value);

	// pockame min 12ns
	write_delay();

	// zapiseme data vytazenim WR na log. 1
	at91_set_gpio_value(TFT_nWR, 1);
}

// -----------------------------------------------------------------------------
static unsigned int read_data(void __iomem * pio_base)
{
	unsigned int data;

	// nastavime signal RS na data
	at91_set_gpio_value(TFT_DC, 1);

	// aktivujeme chipselect
	at91_set_gpio_value(TFT_nCS, 0);

	// vycteme data
	data = read_raw(pio_base);

	// deaktivujeme chipset
	at91_set_gpio_value(TFT_nCS, 1);

	return data;
}

// -----------------------------------------------------------------------------
static void write_data(void __iomem * pio_base, unsigned int data)
{
	// nastavime signal D/C (RS) na data
	at91_set_gpio_value(TFT_DC, 1);

	// aktivujeme chipselect
	at91_set_gpio_value(TFT_nCS, 0);

	// zapiseme data
	write_raw(pio_base, data);

	// deaktivujeme chipset
	at91_set_gpio_value(TFT_nCS, 1);
}

// -----------------------------------------------------------------------------
static void write_cmd(void __iomem * pio_base, unsigned char cmd)
{
	// nastavime signal RS na data
	at91_set_gpio_value(TFT_DC, 0);

	// aktivujeme chipselect
	at91_set_gpio_value(TFT_nCS, 0);

	// zapiseme cmd
	write_raw(pio_base, cmd);

	// deaktivujeme chipset
	at91_set_gpio_value(TFT_nCS, 1);
}

// -----------------------------------------------------------------------------
static void soft_reset(void __iomem * pio_base)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SOFT_RESET);

	// must wait for 5ms at least
	msleep(5);
}

// -----------------------------------------------------------------------------
static void set_pll_mn(void __iomem * pio_base)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_PLL_MN);
	// M = 49, N = 3, Xtal = 8 MHz
	// PLL = Xtal * (M + 1) / (N + 1) = 100 MHz
	write_data(pio_base, 0x31);
	write_data(pio_base, 0x03);
	// potvrdime nastaveni M a N
	write_data(pio_base, 0x04);
}

// -----------------------------------------------------------------------------
static void start_pll(void __iomem * pio_base)
{
	set_bus_write(pio_base);
	// zapneme PLL
	write_cmd(pio_base, CMD_PLL_START);
	write_data(pio_base, 0x01);

	// pockame 100us na stabilizaci PLL
	udelay(100);

	// nastavime PLL jako systemove hodiny
	write_cmd(pio_base, CMD_PLL_START);
	write_data(pio_base, 0x03);
}

// -----------------------------------------------------------------------------
static void set_lshift_freq(void __iomem * pio_base)
{
	// nastavuji pixel clock (rychlost vykreslovani pixelu za sec)
	// defakto pak urcuje spolu s parametry hsync a vsync refresh rate displaye
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_PCLK);
	// LCDC_FPR = 0x4cccc, PLL = 100MHz
	// PCLK = PLL * (LCDC_FPR + 1) / 2^20 = 30 Mhz
	write_data(pio_base, 0x04);
	write_data(pio_base, 0xcc);
	write_data(pio_base, 0xcc);
}

// -----------------------------------------------------------------------------
static void set_lcd_mode(void __iomem * pio_base)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_LCD_MODE);
	// [5]=0: 18bit rozhrani k LCD panelu,
	// [4]=1: enable FRC or dithering, [3]=1: FRC enable,
	// [2]=0: lshift latch on falling edge, [1]=0: hsync active low, [0]=0: vsync active low
	write_data(pio_base, 0x18);
	// TFT mode
	write_data(pio_base, 0x00);
	// 800 pixelu na sirku HDP = 800 - 1
	write_data(pio_base, 0x03);
	write_data(pio_base, 0x1f);
	// 480 pixelu na vysku VDP = 480 - 1
	write_data(pio_base, 0x01);
	write_data(pio_base, 0xdf);
	// RGB pro oboje
	write_data(pio_base, 0x00);
}

// -----------------------------------------------------------------------------
static void set_address_mode(void __iomem * pio_base)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_ADDR_MODE);
	// [7]=0: pages top->bottom, [6]=0: columns left->right, [5]=0: columns normal mode
	// [4]=0: refresh top->bottom, [3]=0: RGB->RGB, [2]=0: refresh left->right,
	// [1]=1: flip horizontal->flipped, [0]=0: flip vertical->normal
	// Pozn.: musim mit obracene framebuffer a display - asi konfigurace panelu
	// INT070ATFT-TS
	write_data(pio_base, 0x02);
}

// -----------------------------------------------------------------------------
static void set_pixel_data_interface(void __iomem * pio_base)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_DATA_INTERFACE);
	// [2:0]=011: sbernice 16bitova, data prenasena v jednom cyklu(RGB565)
	write_data(pio_base, 0x03);
}

// -----------------------------------------------------------------------------
static void set_pixel_format(void __iomem * pio_base)
{
	/* nastavi pocet bitu pouzitych pro vyjadreni barvy jednoho pixelu
	 * tj. nastavuje barevnou hloubku
	 * v ofiko docu je tento prikaz jako reserved, nicmene v dostupnych
	 * zdrojacich je pouzivan dle nasledujici notace:
	 * Display pixel format[6:4] (POR = 000):
	 * 			000 Reserved
	 * 			001 3-bit/pixel
	 * 			010 8-bit/pixel
	 * 			011 12-bit/pixel
	 * 			100 Reserved
	 * 			101 16-bit/pixel
	 * 			110 18-bit/pixel
	 * 			111 24-bit/pixel
	 */
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_PIXEL_FORMAT);
	// [6:4]=101: 16bit/pixel protoze pouzivame 16bit, RGB565 rozhrani
	write_data(pio_base, (0x05 << 4));
}

// -----------------------------------------------------------------------------
static void set_hori_period(void __iomem * pio_base)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_HOR_PERIOD);
	// HT = 0x0434 (1075)
	// celkova doba pro vykresleni radku (hsync pulse, front a back porch a
	// samotne pixely)
	write_data(pio_base, 0x04);
	write_data(pio_base, 0x34);
	// HPS = 0x0002 (2)
	//
	write_data(pio_base, 0x00);
	write_data(pio_base, 0x02);
	// HPW = 0x15 (21)
	write_data(pio_base, 0x15);
	// LPS = 0x000000
	write_data(pio_base, 0x00);
	write_data(pio_base, 0x00);
	write_data(pio_base, 0x00);
}

// -----------------------------------------------------------------------------
static void set_vert_period(void __iomem * pio_base)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_VER_PERIOD);
	// VT = 0x0221 (545)
	write_data(pio_base, 0x02);
	write_data(pio_base, 0x21);
	// VPS = 0x0016 (22)
	write_data(pio_base, 0x00);
	write_data(pio_base, 0x16);
	// VPW = 0x15 (21)
	write_data(pio_base, 0x15);
	// FPS = 0x000000
	write_data(pio_base, 0x00);
	write_data(pio_base, 0x00);
}

// -----------------------------------------------------------------------------
static void set_column_addr(void __iomem * pio_base,
							unsigned int sc, unsigned int ec)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_COLUMN);
	// start column address (SC)
	write_data(pio_base, (sc >> 8));
	write_data(pio_base, sc);
	// end column address (EC)
	write_data(pio_base, (ec >> 8));
	write_data(pio_base, ec);
}

// -----------------------------------------------------------------------------
static void set_page_addr(void __iomem * pio_base,
						  unsigned int sp, unsigned int ep)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_PAGE);
	// start page address (SP)
	write_data(pio_base, (sp >> 8));
	write_data(pio_base, sp);
	// end page address (EP)
	write_data(pio_base, (ep >> 8));
	write_data(pio_base, ep);
}

// -----------------------------------------------------------------------------
static void set_gpio_value(void __iomem * pio_base)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_GPIO_VAL);
	// nastavime log. 1
	write_data(pio_base, 0x0f);
}

// -----------------------------------------------------------------------------
static void set_gpio_conf(void __iomem * pio_base)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_GPIO_CONF);
	// [7:4]=0: GPIO rizeno MCU, [3:0]=1: GPIO jsou vystupy
	write_data(pio_base, 0x0f);
	// [0]=1: GPIO0 je pouzito jako normalni GPIO ne pro rizeni napajeni
	write_data(pio_base, 0x01);
}

// -----------------------------------------------------------------------------
static void set_display_on(void __iomem * pio_base)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_ON_DISPLAY);
}

// -----------------------------------------------------------------------------
static void set_display_off(void __iomem * pio_base)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_BLANK_DISPLAY);
}

// -----------------------------------------------------------------------------
static void set_pwm_conf(void __iomem * pio_base, unsigned char brightness)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_PWM_CONF);

	// PWM frekvence (PWMclk ~ 300Hz)
	// PWMF = PLL / (256^2 * PWMclk) = 100e6 / (256^2 * 300) = 5,0863 ~ 5
	write_data(pio_base, 0x05);
	// modulacni pomer (0x00 (tma) - 0xFF(max. jas))
	write_data(pio_base, brightness);
	// zapneme PWM a nastavime rizeni z MCU (ne pomoci DBC)
	write_data(pio_base, 0x01);
	// max. jas, kdyz to ridi DBC
	write_data(pio_base, 0x00);
	// min. jas, kdyz to ridi DB
	write_data(pio_base, 0x00);
	// krok zmeny jasu - vypnuto
	write_data(pio_base, 0x00);
}

// -----------------------------------------------------------------------------
static void set_dbc_conf(void __iomem * pio_base)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_ABC_DBC_CONF);

	// aktivuji a konfiguruji dynamic backlight control (DBC)
	// [6]=0: DBC manual brightness enable
	// [5]=0: Transition effect disable
	// [3:2]=11: aggressive mode
	// [0]=1: DBC enable
	write_data(pio_base, 0x0d);

	// Pozn.: To ale jeste neznamena, ze PWM bude rizeno DBC!
	// PWM se konfiguruje pred DBC! Viz popisek k prikazu set_dbc_conf
}

// -----------------------------------------------------------------------------
static void set_post_proc(void __iomem * pio_base)
{
	set_bus_write(pio_base);
	write_cmd(pio_base, CMD_SET_POST_PROC);

	// Contrast 0 - 255
	write_data(pio_base, 120);
	// Brightness 0x00 - 0xFF
	write_data(pio_base, 100);
	// Saturation 0x00 - 0xFF
	write_data(pio_base, 100);
	// vypneme post processing obrazu
	write_data(pio_base, 0x00);
}

// -----------------------------------------------------------------------------
static void clear_tft(struct fb_info * into_ptr)
{
	struct int070atft_par_t * par_ptr = into_ptr->par;
	u16 * fb_ptr = par_ptr->vram_ptr;
	u16 * fb_ref_ptr = par_ptr->vram_ref_ptr;
	ulong fb_size = par_ptr->vram_size / sizeof(u16);
	u16 color = 0x0000; // cerna
	uint i;

	// nastavime sbernici na PIOC na vystup
	set_bus_write(par_ptr->pio_base);

	// zapiseme oblast kterou budeme mazat - cely displej dle rozliseni
	set_column_addr(par_ptr->pio_base, 0, into_ptr->var.xres - 1);
	set_page_addr(par_ptr->pio_base, 0, into_ptr->var.yres - 1);

	// zahajime zapis
	write_cmd(par_ptr->pio_base, CMD_WR_MEMSTART);

	// zapiseme barvy jak do obou framebufferu, tak i do displeje
	for ( i = 0; i < fb_size; i++ )
	{
		fb_ptr[i] = fb_ref_ptr[i] = color;

		write_data(par_ptr->pio_base, fb_ptr[i]);
	}
}

// -----------------------------------------------------------------------------
static void init_pio(struct int070atft_par_t * par_ptr)
{
	// vypneme u PIO pro LCD preruseni
	pio_writel((par_ptr->pio_base + PIO_IDR), (TFT_CTRL_PINS_MASK | TFT_DB0_DB15_PIN_MASK));
	// vypneme u PIO pro LCD pullup odpory
	pio_writel((par_ptr->pio_base + PIO_PUDR), (TFT_CTRL_PINS_MASK | TFT_DB0_DB15_PIN_MASK));
	// nastavime ridici signaly na log. 1
	pio_writel((par_ptr->pio_base + PIO_SODR), TFT_CTRL_PINS_MASK);
	// nastavime data na log. 0 (prikaz NOP)
	pio_writel((par_ptr->pio_base + PIO_CODR), TFT_DB0_DB15_PIN_MASK);
	// prepneme PIO pro TFT na vystupni PIO
	pio_writel(par_ptr->pio_base + (PIO_OER), (TFT_CTRL_PINS_MASK | TFT_DB0_DB15_PIN_MASK));
	// pro D0-D17 povolime zapis dat primo pres PIO_ODSR
	// lze tak zapisovat rovnou cele slovo, nicmene prenesou se jen ty bity,
	// ktere maji v masce 1 v PIO_OWER
	pio_writel((par_ptr->pio_base + PIO_OWER), TFT_DB0_DB15_PIN_MASK);
	// prepneme piny pro TFT na funkci GPIO
	pio_writel((par_ptr->pio_base + PIO_PER), (TFT_CTRL_PINS_MASK | TFT_DB0_DB15_PIN_MASK));
}

// -----------------------------------------------------------------------------
static void init_tft(struct int070atft_par_t * par_ptr)
{
	// mozna nahradit zatahanim za HW reset? Nebo ten bude pripojen k NRST MCU?
	soft_reset(par_ptr->pio_base);
	// nastaveni a aktivace PLL jako systemovych hodin
	set_pll_mn(par_ptr->pio_base);
	start_pll(par_ptr->pio_base);
	// znovu SW reset
	soft_reset(par_ptr->pio_base);
	// nastaveni pixel clock - rychlost vykreslovani obrazovych bodu
	set_lshift_freq(par_ptr->pio_base);
	// nastavujeme parametry TFT panelu a zpusob adresace framebuferu
	set_lcd_mode(par_ptr->pio_base);
	set_address_mode(par_ptr->pio_base);
	// nastavujeme sirku datove sbernice a barevnou hloubku/schema
	set_pixel_data_interface(par_ptr->pio_base);
	set_pixel_format(par_ptr->pio_base);
	// nastavujeme hsync a vsync parametry
	set_hori_period(par_ptr->pio_base);
	set_vert_period(par_ptr->pio_base);
	// nastavujeme zobrazovanou plochu
	set_column_addr(par_ptr->pio_base, TFT_SC, TFT_EC);
	set_page_addr(par_ptr->pio_base, TFT_SP, TFT_EP);
	// GPIO piny jako vystupy a do log. 1
	set_gpio_value(par_ptr->pio_base);
	set_gpio_conf(par_ptr->pio_base);
	// prepneme display do stavu DISPLAY
	set_display_on(par_ptr->pio_base);

	// 0 (tma) - 255 (max jas)
	set_pwm_conf(par_ptr->pio_base, 180);
	// PWM tim rizeno neni
	set_dbc_conf(par_ptr->pio_base);
	// nastavime kontrast, jas a saturaci
	set_post_proc(par_ptr->pio_base);
}

// -----------------------------------------------------------------------------
static void * rvmalloc(unsigned long mem_size)
{
	void * mem_ptr = NULL;

	// zarovname pozadovanou velikost na velikost stranky (4kB)
	mem_size = PAGE_ALIGN(mem_size);

	mem_ptr = vmalloc_32(mem_size);
	if ( mem_ptr != NULL )
	{
		u8 * tmp_mem_ptr = mem_ptr;

		// smazeme pamet - zabranime unikum dat do userspace
		memset(mem_ptr, 0x00, mem_size);

		// rezervujeme si pametove stranky alokovane pomoci vmalloc, protoze
		// prostor alokovany pomoci vmalloc budeme mmapovat do userspace
		// tak se muze stat, ze by swapper mohl tyto stranky odswapovat
		// coz je sice na ARMu nepravdepodobny, ale jistota je jistota.
		// Dalsi duvod viz kometar z jedne diskuze:
		/*
		 * 1) So swapper doesn't care if it is the result of kmalloc in kernel.
		 * It just sees accessible from user space bit.
		 *
		 * ad 1) As noted, this bit stops a "race condition" that could otherwise
		 * exist between this code and the swapper. If two pieces of code
		 * (the user, and the kernel) are relying upon a particular physical-page,
		 * that page is now a shared resource between the two for the duration,
		 * and it is necessary to guarantee that the page will not be "stolen"
		 * in the meantime. The bit indicates among other things that this page
		 * is not a candidate for stealing
		 */
		while ( (long) mem_size > 0 )
		{
			SetPageReserved(vmalloc_to_page(tmp_mem_ptr));
			tmp_mem_ptr += PAGE_SIZE;
			mem_size -= PAGE_SIZE;
		}
	}

	return mem_ptr;
}

// -----------------------------------------------------------------------------
static void rvfree(void * mem_ptr, unsigned long mem_size)
{
	if ( mem_ptr != NULL )
	{
		u8 * tmp_mem_ptr = mem_ptr;

		while ( (long) mem_size > 0 )
		{
			ClearPageReserved(vmalloc_to_page(tmp_mem_ptr));
			tmp_mem_ptr += PAGE_SIZE;
			mem_size -= PAGE_SIZE;
		}

		vfree(mem_ptr);
	}
}

// -----------------------------------------------------------------------------
static void refresh_screen_16(struct fb_info * into_ptr)
{
	struct int070atft_par_t * par_ptr = into_ptr->par;
	u16 * fb_ptr = par_ptr->vram_ptr;
	u16 * fb_ref_ptr = par_ptr->vram_ref_ptr;
	ulong fb_size = par_ptr->vram_size / sizeof(u16);
	uint xpos = 0, ypos = 0, xpos_prev = TFT_MAX_COLS + 1;
	uint i;
	bool data_written = false;

	// nastavime sbernici na PIOC na vystup
	set_bus_write(par_ptr->pio_base);

	// porovname a preneseme zmenena data do radice TFT
	for ( i = 0; i < fb_size; i++ )
	{
		if ( fb_ptr[i] != fb_ref_ptr[i] )
		{
			// sousedni pixel na radku?
			xpos = i % into_ptr->var.xres;
			if ( xpos != (xpos_prev + 1) )
			{
				// ne: nastav novou pozici a zapis ji
				ypos = i / into_ptr->var.xres;

				set_column_addr(par_ptr->pio_base, xpos, into_ptr->var.xres - 1);
				set_page_addr(par_ptr->pio_base, ypos, into_ptr->var.yres - 1);

				write_cmd(par_ptr->pio_base, CMD_WR_MEMSTART);
			}
			else
			{
				// ano: rekni radici LCD, ze pokracujeme s dalsim pixelem
				// pozn.: bez tohoto prikazu zustavaji pri pohybu objetu
				// na displeji artefakty - typicky pohybujici se kurzor mysi
				// asi souvisi s puvodnim problemem - chybejici posledni pixel
				// pokud posledni vykreslovany pixel koncil na liche pozici
				// patch, ktery tohle osetroval nepomohl.
				write_cmd(par_ptr->pio_base, CMD_WR_MEM_AUTO);
			}

			write_data(par_ptr->pio_base, fb_ptr[i]);
			data_written = true;

			fb_ref_ptr[i] = fb_ptr[i];
			xpos_prev = xpos;
		}
	}

	// patch: timto je osetreno nezobrazovani posledniho pixelu, kdyz jeho pozice
	// vyjde jako licha (v ramci radku)
	if ( data_written )
	{
		write_cmd(par_ptr->pio_base, CMD_WR_MEMSTART);
	}
}

// -----------------------------------------------------------------------------
static void task_refresh_screen(unsigned long data)
{
	struct fb_info * info_ptr = (struct fb_info *) data;
	struct int070atft_par_t * par_ptr = NULL;

	if ( info_ptr == NULL )
	{
		return;
	}

	par_ptr = info_ptr->par;

	// podle barevne hloubky zavolame odpovidajici fci pro prekresleni displeje
	switch ( info_ptr->var.bits_per_pixel )
	{
		case 16:
			refresh_screen_16(info_ptr);
			break;

		default:
			break;
	}

	// pokud to zakazano, znovunaplanujeme timer
	if ( !atomic_read(&par_ptr->stop_refresh_timer) )
	{
		mod_timer(&par_ptr->refresh_timer,
				  jiffies + msecs_to_jiffies(TFT_REFRESH_PERIOD));
	}
}

// -----------------------------------------------------------------------------
/*
 * Podle stavu HW nastavime struktury fb_var_screeninfo (info_ptr->var) a
 * fb_fix_screeninfo (info->fix), tak aby udaje v fb_info odpovidaly tomu, co je
 * v HW (radici SSD1963).
 *
 * Pozn.: Momentalne je radic SSD1963 inicialiozvan az v tomto ovladaci, takze
 * udaje pouze nakopirujeme ze symbolickych konstant. Zadne vycitani z HW
 * se nekona.
 */
static void __devinit fetch_hw_state(struct fb_info * info_ptr)
{
	struct fb_var_screeninfo * var_ptr = &info_ptr->var;
	struct fb_fix_screeninfo * fix_ptr = &info_ptr->fix;

	// fb_var_screeninfo
	//
	// fyzicke a virtualni rozliseni v pixelech
	var_ptr->xres = TFT_MAX_COLS;
	var_ptr->yres = TFT_MAX_ROWS;
	var_ptr->xres_virtual = TFT_MAX_COLS;
	var_ptr->yres_virtual = TFT_MAX_ROWS;
	var_ptr->xoffset = 0;
	var_ptr->yoffset = 0;
	// pocet bitu barevne hloubky na pixel, pouzivame 16bit schema RGB565
	var_ptr->bits_per_pixel = TFT_MAX_BPP;
	// pouzivame barvy ne odstiny sede
	var_ptr->grayscale = 0;
	// schema RGB565:
	// 5 bitu pro R, 11 bitu posun, little endian = MSB je vlevo
	var_ptr->red.length = 5; var_ptr->red.offset = 11; var_ptr->red.msb_right = 0;
	// 6 bitu pro G, 5 bitu posun, --""--
	var_ptr->green.length = 6; var_ptr->green.offset = 5; var_ptr->green.msb_right = 0;
	// 5 bitu pro G, 0 bitu posun, --""--
	var_ptr->blue.length = 5; var_ptr->blue.offset = 0; var_ptr->blue.msb_right = 0;
	// nebudeme nastaveni aktivovat jen otestujeme, ze by slo nastavit
	var_ptr->activate = FB_ACTIVATE_NOW;

	// fb_fix_screeninfo
	//
	// pocet bytu na radek displeje
	fix_ptr->line_length = var_ptr->xres * var_ptr->bits_per_pixel;
	fix_ptr->line_length /= 8;
}

// -----------------------------------------------------------------------------
static int int070atftfb_mmap(struct fb_info * info_ptr,
							 struct vm_area_struct * vma_ptr)
{
	int retval = 0;
	// pozadavky na mmapovanou pamet z user space
	// Pozor: pozadovana velikost vyjde zarovnana na velikost stranky!
	ulong vma_mem_ptr = vma_ptr->vm_start;
	ulong vma_mmem_size = vma_ptr->vm_end - vma_ptr->vm_start;
	ulong vma_mmem_offset = vma_ptr->vm_pgoff << PAGE_SHIFT;
	// zacatek (posunuty o pozadovany offset) a velikost framebufferu
	// (zarovnane na velikost stranky) - alokovano pomoci vmalloc
	u8 * fb_ptr = (u8 *) (info_ptr->fix.smem_start + vma_mmem_offset);
	ulong fb_size_pg = PAGE_ALIGN(info_ptr->fix.smem_len);
	// velikost framebufferu (alokovane pameti) zarovnane na velikost stranky


	// pozadovana velikost mmapovane pameti nesmi presahnout velikost virtualni
	// pameti alokovane ovladacem
	if ( (vma_mmem_offset + vma_mmem_size) > fb_size_pg )
	{
		retval = -EINVAL;
		goto end;
	}

	// virtualni pamet mmapujeme stranku po strance - stranky vytvarejici
	// virtualni pamet, maji kazda jiny preklad adresy, takze to skutecne musime
	// delat stranku po strance.
	// viz napr. zde: http://www.scs.ch/~frey/linux/memorymap.html
	while ( vma_mmem_size > 0 )
	{
		unsigned long page;

		page = vmalloc_to_pfn(fb_ptr);
		retval = remap_pfn_range(vma_ptr, vma_mem_ptr, page, PAGE_SIZE, PAGE_SHARED);
		if ( retval < 0 )
		{
			retval = -EAGAIN;
			goto end;
		}

		vma_mem_ptr += PAGE_SIZE;
		fb_ptr += PAGE_SIZE;
		if ( vma_mmem_size > PAGE_SIZE )
		{
			vma_mmem_size -= PAGE_SIZE;
		}
		else
		{
			vma_mmem_size = 0;
		}
	}

	// nastavime, ze stranky jsou rezervovane - zabranime jejich odswapovani v
	// user space
	vma_ptr->vm_flags |= VM_RESERVED;

end:
	return retval;
}

// -----------------------------------------------------------------------------
static int int070atftfb_setcolreg(unsigned regno, unsigned red, unsigned green,
			   	   	   	   	   	  unsigned blue, unsigned transp,
			   	   	   	   	   	  struct fb_info * info_ptr)
{
	int retval = 0;

	// pokud je aktivni zobrazovani v odstinech sedi, tak barvy patricne
	// prepocitame
	// grayscale = (0.30 * red) + (0.59 * green) + (0.11 * blue)
	if ( info_ptr->var.grayscale )
	{
		red = green = blue = (19595 * red + 38470 * green + 7471 * blue) >> 16;
	}

	// podle modu provedeme patricne nastaveni patricneho registru
	// v pseudopalete (truecolor) nebo v look up table (lut)
	// v radici LCD/TFT (pseudocolor) - nebylo implementovano.
	// pozn.: pseudopaleta musi byt nastavena kvuli systemove konzoli
	// - ovladac fbcon.
	/*
	 * skeletonfb.c: This structure is used _only_ by fbcon, thus
	 * it only contains 16 entries to match the number of colors supported
	 * by the console.
	 */
	switch ( info_ptr->fix.visual )
	{
		case FB_VISUAL_TRUECOLOR:
		{
			u32 pseudo_val;

			if ( regno >= 16 )
			{
				retval = -EINVAL;
				goto end;
			}

			pseudo_val  = (red   >> 11) << info_ptr->var.red.offset;
			pseudo_val |= (green >> 10) << info_ptr->var.green.offset;
			pseudo_val |= (blue  >> 11) << info_ptr->var.blue.offset;

			((u32 *) info_ptr->pseudo_palette)[regno] = pseudo_val;
		}
		break;

		default:
			retval = -ENOSYS;
			goto end;
			break;
	}

end:
    return retval;
}

// -----------------------------------------------------------------------------
static struct fb_ops int070atftfb_fbops =
{
	.owner			= THIS_MODULE,
	// nas framebuffer je v systemove RAM, takze pouzijeme pro to urcene fce
	.fb_read        = fb_sys_read,
	.fb_write       = fb_sys_write,

	.fb_mmap		= &int070atftfb_mmap,

	// neimplementujeme fci fb_check_var - nas HW ma pevne dany mod, vyssi vrstva
	// zajisti zkopirovani info_ptr->var do var. Predepsane to chovani pro
	// nas pripad
	.fb_check_var	= NULL,
	// neimplementujeme fci fb_set_var - nas HW ma pevne dany mod
	.fb_set_par		= NULL,
	.fb_setcolreg	= &int070atftfb_setcolreg,
	// zatim neimplementovano
	.fb_blank		= NULL,

	// zadnou hw akceleraci nemame, frambufer je navic v RAM, takze pouzijeme
	// pro to urcene fce s sw akceleraci
	.fb_fillrect	= sys_fillrect,
	.fb_copyarea	= sys_copyarea,
	.fb_imageblit	= sys_imageblit,
};

// -----------------------------------------------------------------------------
static void cleanup(struct fb_info * info_ptr, int error)
{
	if ( info_ptr != NULL )
	{
		struct int070atft_par_t * priv_par_ptr =
									(struct int070atft_par_t *) info_ptr->par;

		switch ( error )
		{
			// kompletni uklid pri exitu
			case DrvRemove:
				// zastaveni a zruseni casovace
				atomic_set(&priv_par_ptr->stop_refresh_timer, 1);
				del_timer(&priv_par_ptr->refresh_timer);

				// deregistrace framebufferoveho zarizeni
				unregister_framebuffer(info_ptr);

			case Failed_FbRegister:
				// vypnuti displeje a podsviceni
				set_pwm_conf(priv_par_ptr->pio_base, 0);
				set_display_off(priv_par_ptr->pio_base);

				// uvolnime pamet pro colormapu
				fb_dealloc_cmap(&info_ptr->cmap);

			case Failed_CmapAlloc:
				// uvolnime pamet pro referencni framebuffer
				rvfree(priv_par_ptr->vram_ref_ptr, priv_par_ptr->vram_ref_size);

			case Failed_VmemrefAlloc:
				// uvolnime pamet pro framebuffer
				rvfree(priv_par_ptr->vram_ptr, priv_par_ptr->vram_size);

			case Failed_VmemAlloc:
				// zrusime mapovani I/O pameti do adresoveho prostoru jadra
				iounmap(priv_par_ptr->pio_base);

			case Failed_IoMap:
				// zrusime rezervaci I/O portu
				release_mem_region(AT91_PIOC_BASE, AT91_PIOC_SIZE);

			case Failed_ReqRegion:
				framebuffer_release(info_ptr);

			case Failed_FbAlloc:
				// jeste nic se nealokovalo, takze neni co vracet
				break;

			default:
				break;
		}
	}
}

// -----------------------------------------------------------------------------
static int __devinit int070atftfb_probe(struct platform_device * pdev_ptr)
{
	int retval = 0;
	struct fb_info * info_ptr = NULL;
	struct int070atft_par_t * priv_par_ptr = NULL;
	struct device * dev_ptr = &pdev_ptr->dev;
	struct resource * tmp_region_ptr = NULL;

	dev_info(dev_ptr, "probing INT070ATFT/TS framebuffer driver\n");

	// alokujeme si pamet pro fb_info a zaroven i pamet pro privatni parametry
	// ovladace dostupne pod info_ptr->par
	info_ptr = framebuffer_alloc(sizeof(struct int070atft_par_t), &pdev_ptr->dev);
	if ( info_ptr == NULL )
	{
		dev_err(dev_ptr, "allocating memory for fb_info failed!\n");
		cleanup(info_ptr, Failed_FbAlloc);
		retval = -ENOMEM;
		goto end;
	}

	// void pointer si ulozime do typoveho pointeru a dale pracujeme jen s nim
	priv_par_ptr = (struct int070atft_par_t *) info_ptr->par;

	// nastavime strukturu info jako privatni data do device descriptoru
	platform_set_drvdata(pdev_ptr, info_ptr);

	// rezervujeme si potrebne registry MCU pro komunikaci s TFT
	tmp_region_ptr = request_mem_region(AT91_PIOC_BASE, AT91_PIOC_SIZE,
										INT070ATFT_DEVNAME);
	if ( tmp_region_ptr == NULL )
	{
		dev_err(dev_ptr, "reserving I/O memory failed!\n");
		cleanup(info_ptr, Failed_ReqRegion);
		retval = -EBUSY;
		goto end;
	}

	// premapujeme PIO port do adresoveho prostoru jadra
	priv_par_ptr->pio_base = ioremap_nocache(AT91_PIOC_BASE, AT91_PIOC_SIZE);
	if ( priv_par_ptr->pio_base == NULL )
	{
		dev_err(dev_ptr, "ioremapping I/O registers failed!\n");
		cleanup(info_ptr, Failed_IoMap);
		retval = -ENOMEM;
		goto end;
	}

	// rezervujeme si videopamet - tam se bude kreslit z userspace
	// velikost je max rozliseni x max. pocet bitu na pixel prepocitano na byte
	priv_par_ptr->vram_size = (TFT_MAX_COLS * TFT_MAX_ROWS * TFT_MAX_BPP) / 8;
	priv_par_ptr->vram_ptr = rvmalloc(priv_par_ptr->vram_size);
	if ( priv_par_ptr->vram_ptr == NULL )
	{
		dev_err(dev_ptr, "alloccating framebuffer memory failed!\n");
		cleanup(info_ptr, Failed_VmemAlloc);
		retval = -ENOMEM;
		goto end;
	}

	// rezervujeme si referencni videopamet - pro porovnani co se zmenilo
	// velikost je max rozliseni x max. pocet bitu na pixel prepocitano na byte
	priv_par_ptr->vram_ref_size = (TFT_MAX_COLS * TFT_MAX_ROWS * TFT_MAX_BPP) / 8;
	priv_par_ptr->vram_ref_ptr = rvmalloc(priv_par_ptr->vram_ref_size);
	if ( priv_par_ptr->vram_ref_ptr == NULL )
	{
		dev_err(dev_ptr, "alloccating reference framebuffer memory failed!\n");
		cleanup(info_ptr, Failed_VmemrefAlloc);
		retval = -ENOMEM;
		goto end;
	}

	// ukazatel na framebuffer (alokovanou virtualni pamet) ulozime do info
	// Pozn.: normane bychom tuto pamet premapovali do prostoru jadra jako I/O
	// pamet.  Radic SSD1963 ovsem primo pristupnou I/O pamet nema.
	info_ptr->screen_base = (char __iomem *) priv_par_ptr->vram_ptr;
	info_ptr->screen_size = priv_par_ptr->vram_size;

	// nastavime zbytek clenu fb_info
	info_ptr->pseudo_palette = priv_par_ptr->pseudo_palette;
	info_ptr->fbops = &int070atftfb_fbops;
	// vychozi priznaky + frambufer je v systemove RAM a ne v zarizeni
	// a nemame zadnou HW akceleraci
	info_ptr->flags = FBINFO_DEFAULT | FBINFO_VIRTFB | FBINFO_HWACCEL_NONE;

	// alokujeme pamet pro colormapu
	retval = fb_alloc_cmap(&info_ptr->cmap, 256, 0);
	if ( retval < 0 )
	{
		dev_err(dev_ptr, "allocating memory for cmap failed!\n");
		cleanup(info_ptr, Failed_CmapAlloc);
		goto end;
	}

	// zkopirujeme si do info_ptr->fix vychozi hodnoty ze fb_fix_screeninfo
	info_ptr->fix = int070atftfb_fix;
	// nastavime dalsi cleny info_ptr->fix (sem si uklada ukazatel a delka
	// rezervovane I/O pameti - nikoliv ukazatel na uz preampovanou I/O pamet
	// pomoci ioremap())
	info_ptr->fix.mmio_start = AT91_PIOC_BASE;
	info_ptr->fix.mmio_len = AT91_PIOC_SIZE;
	info_ptr->fix.smem_start = (unsigned long) priv_par_ptr->vram_ptr;
	info_ptr->fix.smem_len = priv_par_ptr->vram_size;

	// inicializace pio a tft
	// normalne je to soucasti inicializace plaftormy,
	// tj. v arch/arm/at91-mach/board...
	init_pio(priv_par_ptr);
	init_tft(priv_par_ptr);

	// nastaveni info_ptr->var a info_ptr->fix dle provoznich parametru HW
	fetch_hw_state(info_ptr);

	// vymazani displeje
	clear_tft(info_ptr);

	// registrace ovladace jako framebuferoveho zarizeni - od ted muzeme
	retval = register_framebuffer(info_ptr);
	if ( retval < 0 )
	{
		dev_err(dev_ptr, "registering framebuffer device failed!\n");
		cleanup(info_ptr, Failed_FbRegister);
		goto end;
	}

	// inicializace a spusteni timeru pro refresh lcd
	// povolime znovuspusteni casovace
	atomic_set(&priv_par_ptr->stop_refresh_timer, 0);
	// nastavime uzivatelskou fci a priv data
	priv_par_ptr->refresh_timer.function = &task_refresh_screen;
	priv_par_ptr->refresh_timer.data = (unsigned long) info_ptr;
	init_timer(&priv_par_ptr->refresh_timer);
	// nastavime periodu a aktivujeme casovac
	priv_par_ptr->refresh_timer.expires =
								jiffies + msecs_to_jiffies(TFT_REFRESH_PERIOD);
	add_timer(&priv_par_ptr->refresh_timer);

end:
	return retval;
}

// -----------------------------------------------------------------------------
static int int070atftfb_remove(struct platform_device * pdev_ptr)
{
	struct fb_info * info_ptr = platform_get_drvdata(pdev_ptr);

	dev_info(&pdev_ptr->dev, "removing INT070ATFT/TS framebuffer driver\n");

	cleanup(info_ptr, DrvRemove);

	return 0;
}

// -----------------------------------------------------------------------------
static struct platform_driver int070atftfb_driver =
{
	.probe		= int070atftfb_probe,
	.remove		= int070atftfb_remove,
	.driver		=
	{
		.name	= INT070ATFT_DEVNAME,
		.owner	= THIS_MODULE,
	},
};

static struct platform_device * int070atftfb_device_ptr;

// -----------------------------------------------------------------------------
static int __init int070atftfb_init(void)
{
	int retval = 0;

	// registruji platform driver
	//
	// Pozn.: normalne, nekde mimo kod driveru, by se melo registrovat i
	// platform device - jakasi abstrakce skutecneho hot/coldplug zarizeni.
	// Normlane se tak deje treba v definici platformy, napr. v nasem pripade v
	// definici vlastnosti modulu zde: arch/arm/mach-at91/board-ucsimply_sam9260
	// zaroven se na tom miste definuji i potrebne I/O zdroje, jako registry,
	// IRQ apod.
	retval = platform_driver_register(&int070atftfb_driver);
	if ( retval == 0 )
	{
		// vytvorime a registruje platform_device, id = -1 (vice zarizeni
		// s timtez jmenem nebude) a predame mu zdroje - v tomto pripade nic
		// nepredavame - potrebne zdroje jsou hardcoded do hlav. souboru a fce probe()
		//
		// touto registraci dojde k prohledani virtualni sbernice platform_bus
		// a obslouzeni zarizeni pomoci odpovidajiciho platform driveru =
		// vola se fce probe(). Shoda device a driveru se urcuje jmenem
		int070atftfb_device_ptr = platform_device_register_simple
														(
															INT070ATFT_DEVNAME,
															-1,
															NULL, 0
														);
		if ( IS_ERR(int070atftfb_device_ptr) )
		{
			platform_driver_unregister(&int070atftfb_driver);
			retval = PTR_ERR(int070atftfb_device_ptr);
		}
	}

	return retval;
}

// -----------------------------------------------------------------------------
static void __exit int070atftfb_exit(void)
{
	platform_device_unregister(int070atftfb_device_ptr);
	platform_driver_unregister(&int070atftfb_driver);
}

// -----------------------------------------------------------------------------
module_init(int070atftfb_init);
module_exit(int070atftfb_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Framebuffer driver for INT070ATFT/INT070ATFT-TS");
MODULE_AUTHOR("Marek Laznicka <marek.laznicka@elvoris.cz>");
MODULE_VERSION("0.1.22");
